from thumbor.config import Config


__version__ = "1.2.5"


Config.define(
    'IMAGE_ENGINE',
    'thumbor.engines.pil',
    'The engine to use for non-video files',
    'Imaging')

Config.define(
    'FFMPEG_ENGINE',
    'thumbor_video_engine.engines.ffmpeg',
    'The engine to use for video files',
    'Video')

Config.define(
    'GIFSICLE_PATH',
    None,
    'The path to the gifsicle binary',
    'Imaging')

Config.define(
    'GIFSICLE_ARGS',
    [],
    'Additional CLI arguments to pass to gifsicle',
    'Imaging')

Config.define(
    'FFMPEG_USE_GIFSICLE_ENGINE',
    False,
    'Equivalent to USE_GIFSICLE_ENGINE, but for the ffmpeg engine',
    'Video')

Config.define(
    'FFMPEG_HANDLE_ANIMATED_GIF',
    True,
    'Whether to process animated gifs with the ffmpeg engine',
    'Video')

Config.define(
    'FFMPEG_GIF_AUTO_WEBP',
    True,
    'Specifies whether animated WebP format should be used automatically if '
    'the source image is an animated gif and the request accepts it (via '
    'Accept header)',
    'Video')

Config.define(
    'FFPROBE_PATH',
    '/usr/local/bin/ffprobe',
    'Path for the ffprobe binary',
    'Video')

Config.define(
    'FFMPEG_H264_TWO_PASS',
    False,
    'Whether to use two-pass encoding for h264 in ffmpeg',
    'Video')

Config.define(
    'FFMPEG_H264_TWO_PASS',
    False,
    'Whether to use two-pass encoding for h264 in ffmpeg',
    'Video')

Config.define(
    'FFMPEG_H264_PRESET',
    None,
    'The -preset flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_LEVEL',
    None,
    'The -level flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_PROFILE',
    None,
    'The -profile:v flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_TUNE',
    None,
    'The -tune flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_CRF',
    None,
    'The -crf flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_VBR',
    None,
    'The average bitrate to be used by ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_MAXRATE',
    None,
    'The -maxrate flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_BUFSIZE',
    None,
    'The -bufsize flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_QMIN',
    None,
    'The -qmin flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H264_QMAX',
    None,
    'The -qmax flag passed to ffmpeg for h264 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_TWO_PASS',
    False,
    'Whether to use two-pass encoding for h265 in ffmpeg',
    'Video')

Config.define(
    'FFMPEG_H265_PRESET',
    None,
    'The -preset flag passed to ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_LEVEL',
    None,
    'The -level flag passed to ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_MAXRATE',
    None,
    'The --vbv-maxrate flag passed to ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_BUFSIZE',
    None,
    'The --vbv-bufsize flag passed to libx265',
    'Video')

Config.define(
    'FFMPEG_H265_CRF_MIN',
    None,
    'The --crf-min flag passed to libx265',
    'Video')

Config.define(
    'FFMPEG_H265_CRF_MAX',
    None,
    'The --crf-max flag passed to libx265',
    'Video')

Config.define(
    'FFMPEG_H265_PROFILE',
    None,
    'The -profile:v flag passed to ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_TUNE',
    None,
    'The -tune flag passed to ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_CRF',
    None,
    'The -crf flag passed to ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_H265_VBR',
    None,
    'The average bitrate to be used by ffmpeg for h265 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_TWO_PASS',
    False,
    'Whether to use two-pass encoding for vp9 in ffmpeg',
    'Video')

Config.define(
    'FFMPEG_VP9_VBR',
    None,
    'The average bitrate to be used by ffmpeg for vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_LOSSLESS',
    False,
    'Whether to use lossless vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_DEADLINE',
    None,
    'The -deadline flag passed to ffmpeg for vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_CRF',
    None,
    'The constant quality (-crf) to use by ffmpeg for vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_CPU_USED',
    None,
    'The -cpu-used flag passed to ffmpeg for vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_ROW_MT',
    False,
    'Whether to enable row-based multithreading (-row-mt 1) for vp9 encoding in ffmpeg',
    'Video')

Config.define(
    'FFMPEG_VP9_MINRATE',
    None,
    'The -minrate flag passed to ffmpeg for vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_VP9_MAXRATE',
    None,
    'The -maxrate flag passed to ffmpeg for vp9 encoding',
    'Video')

Config.define(
    'FFMPEG_WEBP_LOSSLESS',
    False,
    '(-lossless) Enables/disables use of lossless mode. libwebp default is False',
    'Video')

Config.define(
    'FFMPEG_WEBP_COMPRESSION_LEVEL',
    None,
    '(-compression_level) range 0-6, default 4. Higher values give better '
    'quality but slower speed. For lossless, it controls the size/speed trade-off',
    'Video')

Config.define(
    'FFMPEG_WEBP_QSCALE',
    None,
    '(-qscale) For lossy encoding, controls quality 0 to 100. For lossless, '
    'controls cpu and time spent compressing. libwebp built-in default 75.',
    'Video')

Config.define(
    'FFMPEG_WEBP_PRESET',
    None,
    '(-preset) Configuration preset. Consult ffmpeg libwebp codec documentation '
    'for more information.',
    'Video')

Config.define(
    'FFMPEG_GIF_AUTO_H264',
    False,
    'Specifies whether an H264 mp4 should be returned automatically if '
    'the source image is an animated gif and the request accepts it (via '
    'Accept: video/* header)',
    'Video')

Config.define(
    'FFMPEG_GIF_AUTO_H265',
    False,
    'Specifies whether an H265 mp4 should be returned automatically if '
    'the source image is an animated gif and the request accepts it (via '
    'Accept: video/* header)',
    'Video')


Config.define(
    'FFMPEG_PRESERVE_AUDIO',
    False,
    'If True, thumbor-video-engine will not strip the audio track. Defaults to '
    'False',
    'Video')



import six


__all__ = ['filter_retval']


# Thumbor 7.0+ requires filters to return an async_generator; by adding
# ``return filter_retval()`` to the end of our filters, we can achieve
# compatibility with Python 2 + Thumbor < 7 and Python 3 + Thumbor 7+
def filter_retval():
    pass


if six.PY3:
    exec("async def filter_retval(): pass")



from os.path import join


class BaseStorage(object):
    @property
    def is_auto_webp(self):
        return (
            self.context.config.AUTO_WEBP and
            hasattr(self.context, 'request') and
            getattr(self.context.request, "accepts_webp", False))

    @property
    def is_auto_video(self):
        auto_video = (
            self.context.config.FFMPEG_GIF_AUTO_H264 or
            self.context.config.FFMPEG_GIF_AUTO_H265)
        return (
            auto_video and
            hasattr(self.context, 'request') and
            getattr(self.context.request, "accepts_video", False))

    def get_auto_path_component(self):
        if self.is_auto_video:
            return "mp4"
        elif self.is_auto_webp:
            return "webp"



import pytest

from thumbor_video_engine.engines.video import BaseEngine


@pytest.mark.parametrize('fname,mime_type', [
    ('hotdog.mp4', 'video/mp4'),
    ('hotdog.h265.mp4', 'video/mp4'),
    ('corrupt2.mp4', None),
])
def test_get_mimetype_monkeypatch(storage_path, fname, mime_type):
    with open("%s/%s" % (storage_path, fname), mode='rb') as f:
        buf = f.read()
    assert BaseEngine.get_mimetype(buf) == mime_type



from thumbor.filters import BaseFilter, filter_method, PHASE_PRE_LOAD
import thumbor.filters.format
from thumbor.utils import logger

from thumbor_video_engine.compat import filter_retval


ALLOWED_FORMATS = list(thumbor.filters.format.ALLOWED_FORMATS) + [
    "webm",
    "mp4",
    "hevc",
    "h264",
    "h265",
    "vp9",
]


class Filter(BaseFilter):
    phase = PHASE_PRE_LOAD

    @filter_method(BaseFilter.String)
    def format(self, format):
        logger.debug("Setting format to %s" % format)
        if format.lower() not in ALLOWED_FORMATS:
            logger.warning("Format not allowed: %s" % format.lower())
            self.context.request.format = None
        else:
            logger.debug("Format specified: %s" % format.lower())
            self.context.request.format = format.lower()

        return filter_retval()



from contextlib import contextmanager

import pytest

import thumbor_video_engine.engines.ffmpeg
import thumbor_video_engine.utils
from thumbor_video_engine.engines.ffmpeg import Engine


@pytest.fixture
def config(config):
    config.ENGINE = 'thumbor_video_engine.engines.ffmpeg'
    return config


@pytest.fixture
def mock_engine(context, mp4_buffer, mocker):
    class MockNamedTemporaryFile(mocker.Mock):
        def __init__(self, suffix=None, **kwargs):
            super(MockNamedTemporaryFile, self).__init__(**kwargs)
            suffix = suffix or ''
            self.name = '/tmp/tempfile%s' % suffix

    engine = Engine(context)

    @contextmanager
    def mock_named_tmp_file(suffix=None, **kwargs):
        suffix = suffix or ''
        yield '/tmp/tempfile%s' % suffix

    mocker.patch.object(thumbor_video_engine.engines.ffmpeg, 'ffprobe',
        return_value={'width': 200, 'height': 150, 'duration': '1.261000'})
    mocker.patch.object(thumbor_video_engine.engines.ffmpeg, 'named_tmp_file',
        wraps=mock_named_tmp_file)
    mocker.patch.object(thumbor_video_engine.engines.ffmpeg, 'open',
        new_callable=mocker.mock_open())
    mocker.patch.object(engine, 'run_cmd', return_value=mp4_buffer)

    engine.load(mp4_buffer, '.mp4')

    return engine


@pytest.fixture
def std_flags():
    return [
        '-an', '-pix_fmt', 'yuv420p', '-movflags', 'faststart',
    ]


@pytest.fixture
def std_h264_flags(ffmpeg_path, std_flags):
    return [
        ffmpeg_path, '-hide_banner', '-i', '/tmp/tempfile.mp4',
        '-c:v', 'libx264'
    ] + std_flags + ['-f', 'mp4']


@pytest.fixture
def std_h265_flags(ffmpeg_path, std_flags):
    return [
        ffmpeg_path, '-hide_banner', '-i', '/tmp/tempfile.mp4',
        '-c:v', 'hevc', '-tag:v', 'hvc1',
    ] + std_flags + ['-f', 'mp4']


@pytest.fixture
def std_vp9_flags(ffmpeg_path, std_flags):
    return [
        ffmpeg_path, '-hide_banner', '-i', '/tmp/tempfile.mp4',
        '-c:v', 'libvpx-vp9', '-loop', '0',
    ] + std_flags + ['-f', 'webm']


@pytest.fixture
def std_webp_flags(ffmpeg_path, std_flags):
    return [
        ffmpeg_path, '-hide_banner', '-i', '/tmp/tempfile.mp4', '-loop', '0',
    ] + std_flags + ['-f', 'webp']


def test_h264_two_pass(mock_engine, std_h264_flags, mocker):
    mock_engine.context.request.format = 'h264'
    mock_engine.context.config.FFMPEG_H264_TWO_PASS = True

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_h264_flags + [
            '-pass', '1', '-passlogfile', '/tmp/tempfile.log', '-y', '/dev/null']),
        mocker.call(std_h264_flags + [
            '-pass', '2', '-passlogfile', '/tmp/tempfile.log', '-y', '/tmp/tempfile.mp4'])]


def test_h265_two_pass(mock_engine, std_h265_flags, mocker):
    mock_engine.context.request.format = 'h265'
    mock_engine.context.config.FFMPEG_H265_TWO_PASS = True

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_h265_flags + [
            '-x265-params', 'pass=1:stats=/tmp/tempfile.log', '-y', '/dev/null']),
        mocker.call(std_h265_flags + [
            '-x265-params', 'pass=2:stats=/tmp/tempfile.log', '-y', '/tmp/tempfile.mp4'])]


def test_h265_two_pass_with_x265_param(mock_engine, std_h265_flags, mocker):
    mock_engine.context.request.format = 'h265'
    mock_engine.context.config.FFMPEG_H265_TWO_PASS = True
    mock_engine.context.config.FFMPEG_H265_CRF_MAX = 16

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_h265_flags + [
            '-x265-params', 'crf-max=16:pass=1:stats=/tmp/tempfile.log',
            '-y', '/dev/null']),
        mocker.call(std_h265_flags + [
            '-x265-params', 'crf-max=16:pass=2:stats=/tmp/tempfile.log',
            '-y', '/tmp/tempfile.mp4'])]


def test_vp9_two_pass(mock_engine, std_vp9_flags, mocker):
    mock_engine.context.request.format = 'vp9'
    mock_engine.context.config.FFMPEG_VP9_TWO_PASS = True

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_vp9_flags + [
            '-pass', '1', '-passlogfile', '/tmp/tempfile.log', '-y', '/dev/null']),
        mocker.call(std_vp9_flags + [
            '-pass', '2', '-passlogfile', '/tmp/tempfile.log', '-y', '/tmp/tempfile.webm'])]


@pytest.mark.parametrize("config_key,config_val,expected", [
    ('FFMPEG_H264_CRF', '22', ['-crf', '22']),
    ('FFMPEG_H264_VBR', '1M', ['-b:v', '1M']),
    ('FFMPEG_H264_PROFILE', 'high', ['-profile:v', 'high']),
    ('FFMPEG_H264_PRESET', 'medium', ['-preset', 'medium']),
    ('FFMPEG_H264_LEVEL', '3.0', ['-level', '3.0']),
    ('FFMPEG_H264_TUNE', 'film', ['-tune', 'film']),
    ('FFMPEG_H264_MAXRATE', '2M', ['-maxrate', '2M']),
    ('FFMPEG_H264_BUFSIZE', '4M', ['-bufsize', '4M']),
    ('FFMPEG_H264_QMIN', '16', ['-qmin', '16']),
    ('FFMPEG_H264_QMAX', '26', ['-qmax', '26']),
])
def test_h264_config(config_key, config_val, expected, mock_engine, std_h264_flags, mocker):
    mock_engine.context.request.format = 'h264'
    setattr(mock_engine.context.config, config_key, config_val)

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_h264_flags + expected + ['-y', '/tmp/tempfile.mp4']),
    ]


@pytest.mark.parametrize("config_key,config_val,expected", [
    ('FFMPEG_H265_VBR', '1M', ['-b:v', '1M']),
    ('FFMPEG_H265_CRF', '22', ['-crf', '22']),
    ('FFMPEG_H265_PROFILE', 'high', ['-profile:v', 'high']),
    ('FFMPEG_H265_PRESET', 'medium', ['-preset', 'medium']),
    ('FFMPEG_H265_TUNE', 'film', ['-tune', 'film']),
    ('FFMPEG_H265_MAXRATE', '2M', ['-x265-params', 'vbv-maxrate=2M']),
    ('FFMPEG_H265_BUFSIZE', '4M', ['-x265-params', 'vbv-bufsize=4M']),
    ('FFMPEG_H265_CRF_MIN', '16', ['-x265-params', 'crf-min=16']),
    ('FFMPEG_H265_CRF_MAX', '26', ['-x265-params', 'crf-max=26']),
])
def test_h265_config(config_key, config_val, expected, mock_engine, std_h265_flags, mocker):
    mock_engine.context.request.format = 'h265'
    setattr(mock_engine.context.config, config_key, config_val)

    mock_engine.read('.mp4', quality=80)

    if '-x265-params' not in expected:
        expected += ['-x265-params', '']

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_h265_flags + expected + ['-y', '/tmp/tempfile.mp4']),
    ]


@pytest.mark.parametrize("config_key,config_val,expected", [
    ('FFMPEG_VP9_CRF', '22', ['-crf', '22']),
    ('FFMPEG_VP9_VBR', '1M', ['-b:v', '1M']),
    ('FFMPEG_VP9_DEADLINE', 'high', ['-deadline', 'high']),
    ('FFMPEG_VP9_CPU_USED', '4', ['-cpu-used', '4']),
    ('FFMPEG_VP9_ROW_MT', True, ['-row-mt', '1']),
    ('FFMPEG_VP9_LOSSLESS', True, ['-lossless', '1']),
    ('FFMPEG_VP9_MAXRATE', '4M', ['-maxrate', '4M']),
    ('FFMPEG_VP9_MINRATE', '1M', ['-minrate', '1M']),
])
def test_vp9_config(config_key, config_val, expected, mock_engine, std_vp9_flags, mocker):
    mock_engine.context.request.format = 'vp9'
    setattr(mock_engine.context.config, config_key, config_val)

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_vp9_flags + expected + ['-y', '/tmp/tempfile.webm']),
    ]


@pytest.mark.parametrize("config_key,config_val,expected", [
    ('FFMPEG_WEBP_LOSSLESS', True, ['-lossless', '1']),
    ('FFMPEG_WEBP_COMPRESSION_LEVEL', '4', ['-compression_level', '4']),
    ('FFMPEG_WEBP_QSCALE', '75', ['-qscale', '75']),
    ('FFMPEG_WEBP_PRESET', 'drawing', ['-preset', 'drawing']),
])
def test_webp_config(config_key, config_val, expected, mock_engine, std_webp_flags, mocker):
    mock_engine.context.request.format = 'webp'
    setattr(mock_engine.context.config, config_key, config_val)

    if config_key == 'FFMPEG_WEBP_LOSSLESS':
        # Swap out pix_fmt for lossless webp
        changes = {'yuv420p': 'rgba'}
        std_webp_flags = [changes.get(f, f) for f in std_webp_flags]

    mock_engine.read('.mp4', quality=80)

    assert mock_engine.run_cmd.mock_calls == [
        mocker.call(std_webp_flags + expected + ['-y', '/tmp/tempfile.webp']),
    ]



from contextlib import contextmanager
from io import BytesIO
import os
import shutil
from struct import unpack
from tempfile import NamedTemporaryFile, mkdtemp

from PIL import Image


@contextmanager
def named_tmp_file(data=None, extension=None, **kwargs):
    kwargs.setdefault('delete', False)
    if extension is not None:
        kwargs.setdefault('suffix', extension)
    f = NamedTemporaryFile(**kwargs)
    try:
        if data:
            f.write(data)
        f.close()
        yield f.name
    finally:
        os.unlink(f.name)


@contextmanager
def make_tmp_dir():
    try:
        tmp_dir = mkdtemp()
        yield tmp_dir
    finally:
        shutil.rmtree(tmp_dir, ignore_errors=True)


def is_mp4(buf):
    if buf[4:8] != b'ftyp':
        return False
    (ftyp_box_len,) = unpack('>L', buf[:4])
    if not (20 <= ftyp_box_len <= 256) or (ftyp_box_len % 4) != 0:
        return False
    major_brand = unpack('4s', buf[8:12])
    compat_brand_len = ftyp_box_len - 16
    fmt = '4s' * (compat_brand_len // 4)
    compat_brands = unpack(fmt, buf[16:ftyp_box_len])
    all_brands = set(major_brand + compat_brands)
    return bool(all_brands & {b'isom', b'avc1', b'iso2', b'mp41', b'mp42'})


def is_qt(buf):
    return buf[4:12] == b'ftypqt  '


def has_transparency(im):
    if 'A' in im.mode or 'transparency' in im.info:
        # If the image has alpha channel, we check for any pixels that are not opaque (255)
        return min(im.convert('RGBA').getchannel('A').getextrema()) < 255
    else:
        return False


def is_animated(buffer):
    im = Image.open(BytesIO(buffer))
    return getattr(im, 'is_animated', False)


def ord_compat(val):
    if isinstance(val, int):
        return val
    else:
        return ord(val)


def is_animated_gif(buffer):
    if buffer[:6] not in [b"GIF87a", b"GIF89a"]:
        return False
    i = 10  # skip header
    frames = 0

    def skip_color_table(i, flags):
        if flags & 0x80:
            i += 3 << ((flags & 7) + 1)
        return i

    flags = ord_compat(buffer[i])

    i = skip_color_table(i + 3, flags)
    while frames < 2:
        block = buffer[i]
        i += 1
        if block in (b'\x3B', 0x3B):
            break
        if block in (b'\x21', 0x21):
            i += 1
        elif block in (b'\x2C', 0x2C):
            frames += 1
            i += 8
            i = skip_color_table(i + 1, ord_compat(buffer[i]))
            i += 1
        else:
            return False
        while True:
            j = ord_compat(buffer[i])
            i += 1
            if not j:
                break
            i += j
    return frames > 1



from thumbor.engines import BaseEngine
from thumbor.utils import logger

from thumbor_video_engine.utils import (
    named_tmp_file, is_mp4, is_qt, is_animated, is_animated_gif)


def patch_baseengine_get_mimetype():
    """
    Monkey-patch BaseEngine.get_mimetype() to recognize all mp4 files as video/mp4
    """
    orig_get_mimetype = BaseEngine.get_mimetype

    def get_mimetype(cls, buffer):
        mimetype = orig_get_mimetype(buffer)
        if mimetype is not None:
            return mimetype
        elif is_qt(buffer):
            return 'video/quicktime'
        elif is_mp4(buffer):
            return 'video/mp4'

    BaseEngine.get_mimetype = classmethod(get_mimetype)


patch_baseengine_get_mimetype()


class Engine(object):
    """An engine that dispatches video files to ffmpeg, and others to PIL"""

    def __init__(self, context):
        self.engine = None
        self.context = context
        self.ffmpeg_handle_animated_gif = context.config.FFMPEG_HANDLE_ANIMATED_GIF
        self.ffmpeg_handle_animated_webp = True
        self.use_gif_engine = context.config.FFMPEG_USE_GIFSICLE_ENGINE

    @property
    def image_engine(self):
        self.context.modules.importer.import_item('IMAGE_ENGINE', 'Engine')
        return self.context.modules.importer.image_engine(self.context)

    @property
    def ffmpeg_engine(self):
        if not hasattr(self.context.modules, 'ffmpeg_engine'):
            # Instantiate the video engine class from the config (default is
            # thumbor_video_engine.engines.ffmpeg)
            self.context.modules.importer.import_item('FFMPEG_ENGINE', 'Engine')
            self.context.modules.ffmpeg_engine = (
                self.context.modules.importer.ffmpeg_engine(self.context))
        return self.context.modules.ffmpeg_engine

    def get_engine(self, buffer, extension):
        mime = BaseEngine.get_mimetype(buffer)

        is_gif = extension == '.gif'
        is_webp = extension == '.webp'
        accepts_video = getattr(self.context.request, "accepts_video", False)
        accepts_webp = self.context.request.accepts_webp

        if is_webp and self.ffmpeg_handle_animated_webp and is_animated(buffer):
            return self.ffmpeg_engine
        elif is_gif and self.ffmpeg_handle_animated_gif and is_animated_gif(buffer):
            if self.context.config.FFMPEG_GIF_AUTO_H265:
                self.context.request.should_vary = True
                if accepts_video:
                    logger.debug("FFMPEG_GIF_AUTO_H265 setting format to h264")
                    self.context.request.format = 'h265'
            elif self.context.config.FFMPEG_GIF_AUTO_H264:
                self.context.request.should_vary = True
                if accepts_video:
                    logger.debug("FFMPEG_GIF_AUTO_H264 setting format to h264")
                    self.context.request.format = 'h264'
            elif self.context.config.FFMPEG_GIF_AUTO_WEBP:
                self.context.request.should_vary = True
                if accepts_webp:
                    logger.debug("FFMPEG_GIF_AUTO_WEBP setting format to webp")
                    self.context.request.format = 'webp'
            return self.ffmpeg_engine
        elif is_gif and self.use_gif_engine:
            return self.context.modules.gif_engine
        elif mime.startswith('video/'):
            return self.ffmpeg_engine
        else:
            return self.image_engine

    def load(self, buffer, extension):
        self.engine = self.get_engine(buffer, extension)
        if self.context.request.format and not self.context.request.filters:
            # RequestParameters.filters is an empty list when none are in the url,
            # and ImagingHandler._write_results_to_client assumes that if
            # context.request.format is set then it came from the format filter.
            # Since we set the format in the engine this causes a TypeError,
            # so we need to ensure that it is a string here.
            self.context.request.filters = ""
        logger.debug("Set engine to %s (extension %s)" % (
            type(self.engine).__module__, extension))
        still_frame_pos = getattr(self.context.request, 'still_position', None)
        # Are we requesting a still frame?
        if self.engine is self.ffmpeg_engine and still_frame_pos:
            with named_tmp_file(data=buffer, suffix=extension) as src_file:
                buffer = self.ffmpeg_engine.run_ffmpeg(
                    src_file, 'png', ['-ss', still_frame_pos, '-frames:v', '1'])
                self.engine = self.image_engine
                extension = '.png'
                if not self.context.request.format:
                    self.context.request.format = 'jpg'

        # Change the default extension if we're transcoding video
        if self.engine is self.ffmpeg_engine and extension == ".jpg":
            extension = ".mp4"

        self.extension = extension
        self.engine.load(buffer, extension)

    def is_multiple(self):
        return False

    def cleanup(self):
        pass

    def __getattribute__(self, attr):
        # For compatibility with some thumbor filters, return the image_engine
        # or video_engine class, when appropriate, for self.__class__
        if attr == "__class__" and self.__dict__.get("engine"):
            return self.__dict__["engine"].__class__
        return object.__getattribute__(self, attr)

    def __getattr__(self, attr):
        if not self.__dict__.get('engine'):
            raise AttributeError("'Engine' object has no attribute '%s'" % attr)
        return getattr(self.engine, attr)

    def __setattr__(self, attr, value):
        if attr in ('engine', 'ffmpeg_handle_animated_gif', 'use_gif_engine'):
            self.__dict__[attr] = value
        elif attr in ('context', 'extension'):
            self.__dict__[attr] = value
            if self.engine:
                setattr(self.engine, attr, value)
        elif self.engine:
            setattr(self.engine, attr, value)
        else:
            self.__dict__[attr] = value



import pytest


from thumbor_video_engine.utils import is_mp4, is_animated_gif


@pytest.mark.parametrize('bool_val,buf', [
    (True, b'\x00\x00\x00\x18ftypmp42\x00\x00\x00\x00mp42mp41'),
    (True, b'\x00\x00\x00\x1cftypisom\x00\x00\x02\x00isomiso2mp41'),
    (True, b'\x00\x00\x00\x20ftypisom\x00\x00\x02\x00isomiso2avc1mp41'),
    (False, b'\x00\x00\x00\x20XXXXisom\x00\x00\x02\x00isomiso2avc1mp41'),
    (False, b'\x00\x00\x00\x20ftypeaaaa\x00\x00\x02\x00bbbbccccddddeeee'),
    (False, b'\x00\x00\x00\x1bftypisom\x00\x00\x02\x00isomiso2mp4'),
    (False, b'\x00\x00\x00\x08ftypisom'),
])
def test_is_mp4(bool_val, buf):
    assert is_mp4(buf) is bool_val


@pytest.mark.parametrize('buf', [
    b'',
    b'GIF89a\xc8\x00\x96\x00\x00\x00\x00\x00',
    b'GIF89a\x01\x00\x01\x00\x00\xff\x00,\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x00;',
])
def test_is_animated_gif_false(buf):
    assert is_animated_gif(buf) is False


def test_is_animated_gif_true(storage_path):
    with open("%s/hotdog.gif" % storage_path, mode="rb") as f:
        im_bytes = f.read()
    assert is_animated_gif(im_bytes) is True



import pytest

from thumbor_video_engine.exceptions import FFmpegError
import thumbor_video_engine.ffprobe
from thumbor_video_engine.ffprobe import ffprobe


def test_ffprobe_path_which(mocker, monkeypatch, mp4_buffer):
    monkeypatch.setattr(thumbor_video_engine.ffprobe, 'FFPROBE_PATH', None)
    mocker.patch.object(
        thumbor_video_engine.ffprobe, 'which', return_value='/opt/bin/ffprobe')
    mocker.patch.object(
        thumbor_video_engine.ffprobe, 'Popen', side_effect=FFmpegError)
    with pytest.raises(FFmpegError):
        ffprobe(mp4_buffer)
    assert thumbor_video_engine.ffprobe.which.mock_calls == [
        mocker.call('ffprobe')
    ]
    assert thumbor_video_engine.ffprobe.FFPROBE_PATH == '/opt/bin/ffprobe'


def test_ffprobe_path_not_found(mocker, monkeypatch, mp4_buffer):
    monkeypatch.setattr(thumbor_video_engine.ffprobe, 'FFPROBE_PATH', None)
    mocker.patch.object(
        thumbor_video_engine.ffprobe, 'which', return_value=None)
    with pytest.raises(FFmpegError) as exc:
        ffprobe(mp4_buffer)
    assert str(exc.value) == 'Could not find ffprobe executable'


def test_ffprobe_error_code(storage_path):
    with open("%s/corrupt2.mp4" % storage_path, mode="rb") as f:
        corrupt_buf = f.read()
    with pytest.raises(FFmpegError) as exc:
        ffprobe(corrupt_buf)
    assert str(exc.value) == "Invalid data found when processing input (-1094995529)"


@pytest.mark.parametrize("stdout", ['FOO', '[]', '{}'])
def test_ffprobe_invalid_data(mocker, mp4_buffer, stdout):
    mock_proc = mocker.Mock(**{'communicate.return_value': (stdout, '')})
    mocker.patch.object(
        thumbor_video_engine.ffprobe, 'Popen', return_value=mock_proc)
    with pytest.raises(FFmpegError) as exc:
        ffprobe(mp4_buffer)
    assert str(exc.value) == 'ffprobe returned invalid data'


def test_ffprobe_flat(storage_path):
    with open("%s/hotdog.mp4" % storage_path, mode="rb") as f:
        aac_data = f.read()
    ret = ffprobe(aac_data, flat=False)
    assert set(ret.keys()) == {'streams', 'format'}


def test_ffprobe_missing_video_stream(storage_path):
    with open("%s/tearing-me-apart.m4a" % storage_path, mode="rb") as f:
        aac_data = f.read()
    with pytest.raises(FFmpegError) as exc:
        ffprobe(aac_data)
    assert str(exc.value) == 'File is missing a video stream'



from datetime import datetime, timezone
from deprecated import deprecated
from hashlib import sha1
from urllib.parse import unquote
from os.path import join

from thumbor.config import Config
from thumbor.engines import BaseEngine
from thumbor.result_storages import ResultStorageResult
from thumbor.utils import logger
import thumbor_aws.result_storage
from thumbor_aws.utils import normalize_path
from .base import BaseStorage

Config.define(
    "TC_AWS_RANDOMIZE_KEYS", False, "Randomize S3 bucket keys", "tc_aws Compatibility"
)
Config.define(
    "TC_AWS_ROOT_IMAGE_NAME",
    "",
    "When resizing a URL that ends in a slash, what should the corresponding cache key be?",
    "tc_aws Compatibility",
)


class Storage(BaseStorage, thumbor_aws.result_storage.Storage):
    @property
    def prefix(self):
        auto_component = self.get_auto_path_component()
        if auto_component:
            return f"{self.root_path}/{auto_component}".lstrip("/")
        else:
            return self.root_path.lstrip("/")

    def normalize_path(self, path):
        if not self.context.config.THUMBOR_AWS_RUN_IN_COMPATIBILITY_MODE:
            return normalize_path(self.prefix, path)

        segments = [path.lstrip("/")]

        root_path = self.context.config.TC_AWS_RESULT_STORAGE_ROOT_PATH

        if root_path:
            segments.insert(0, root_path.lstrip("/"))

        auto_component = self.get_auto_path_component()
        if auto_component:
            segments.append(auto_component)

        if self.context.config.TC_AWS_RANDOMIZE_KEYS:
            segments.insert(0, self._generate_digest(segments))

        normalized_path = join(*segments)
        if normalized_path.endswith("/"):
            normalized_path += self.context.config.TC_AWS_ROOT_IMAGE_NAME

        return unquote(normalized_path)

    def _generate_digest(self, segments):
        return sha1(".".join(segments).encode("utf-8")).hexdigest()

    async def put(self, image_bytes: bytes) -> str:
        file_abspath = self.normalize_path(self.context.request.url)
        logger.debug("[RESULT_STORAGE] putting at %s", file_abspath)
        content_type = BaseEngine.get_mimetype(image_bytes)
        response = await self.upload(
            file_abspath,
            image_bytes,
            content_type,
            self.context.config.AWS_DEFAULT_LOCATION,
        )
        logger.info("[RESULT_STORAGE] Image uploaded successfully to %s", file_abspath)
        return response

    async def get(self) -> ResultStorageResult:
        path = self.context.request.url
        file_abspath = self.normalize_path(path)

        logger.debug("[RESULT_STORAGE] getting from %s", file_abspath)

        exists = await self.object_exists(file_abspath)
        if not exists:
            logger.debug("[RESULT_STORAGE] image not found at %s", file_abspath)
            return None

        status, body, last_modified = await self.get_data(
            self.bucket_name, file_abspath
        )

        if status != 200 or self._is_expired(last_modified):
            logger.debug(
                "[RESULT_STORAGE] cached image has expired (status %s)", status
            )
            return None

        logger.info(
            "[RESULT_STORAGE] Image retrieved successfully at %s.",
            file_abspath,
        )

        return ResultStorageResult(
            buffer=body,
            metadata={
                "LastModified": last_modified.replace(tzinfo=timezone.utc),
                "ContentLength": len(body),
                "ContentType": BaseEngine.get_mimetype(body),
            },
        )

    @deprecated(version="7.0.0", reason="Use result's last_modified instead")
    async def last_updated(  # pylint: disable=invalid-overridden-method
        self,
    ) -> datetime:
        path = self.context.request.url
        file_abspath = self.normalize_path(path)
        logger.debug("[RESULT_STORAGE] getting from %s", file_abspath)

        response = await self.get_object_metadata(file_abspath)
        return datetime.strptime(
            response["ResponseMetadata"]["HTTPHeaders"]["last-modified"],
            "%a, %d %b %Y %H:%M:%S %Z",
        )



import re
import sys

import tornado.gen as gen

import thumbor.app
from thumbor.handlers.imaging import ImagingHandler
from thumbor.result_storages import ResultStorageResult
from thumbor_video_engine.utils import is_animated, is_animated_gif


class VideoEngineImagingHandler(ImagingHandler):
    def _override_write_results_to_client(self, results, content_type):
        is_gif = content_type == 'image/gif'
        is_webp = content_type == 'image/webp'
        is_video = content_type.startswith('video/')
        accepts_video = getattr(self.context.request, "accepts_video", False)
        accepts_webp = self.context.request.accepts_webp
        auto_gif_video = (
            self.context.config.FFMPEG_GIF_AUTO_H264 or
            self.context.config.FFMPEG_GIF_AUTO_H265)
        auto_gif_webp = (
            self.context.config.FFMPEG_GIF_AUTO_WEBP and
            self.context.config.AUTO_WEBP)

        # If the result is from result_storage, we need to determine whether
        # we should Vary the result because of auto-gif-conversion settings
        if isinstance(results, ResultStorageResult):
            buf = results.buffer
            if is_webp and is_animated(buf) and accepts_webp and auto_gif_webp:
                self.context.request.should_vary = True
            elif is_video and accepts_video and auto_gif_video:
                self.context.request.should_vary = True
            elif is_gif and is_animated_gif(buf):
                if auto_gif_video and not accepts_video:
                    self.context.request.should_vary = True
                elif auto_gif_webp and not accepts_webp:
                    self.context.request.should_vary = True

        # RequestParameters.filters is an empty list when none are in the url,
        # and ImagingHandler._write_results_to_client assumes that if
        # context.request.format is set then it came from the format filter.
        # Since we set the format in the engine this causes a TypeError,
        # so we need to ensure that it is a string here.
        if not self.context.request.filters:
            self.context.request.filters = ""
        # output format is not requested via format filter
        should_vary = not (
            # format is supported by filter
            self.context.request.format and
            # filter is in request
            bool(re.search(r"format\([^)]+\)", self.context.request.filters)))

        if should_vary and getattr(self.context.request, "should_vary", False):
            self.set_header('Vary', 'Accept')

    if sys.version_info[0] == 2:
        def _write_results_to_client(self, results, content_type):
            self._override_write_results_to_client(results, content_type)
            super(VideoEngineImagingHandler, self)._write_results_to_client(results, content_type)
    else:
        exec("\n".join([
            "async def _write_results_to_client(self, results, content_type):",
            "    self._override_write_results_to_client(results, content_type)",
            "    await ImagingHandler._write_results_to_client(self, results, content_type)",
        ]))

    def _override_execute_image_operations(self):
        self.context.request.accepts_video = (
            'video/' in self.request.headers.get('Accept', ''))

    if sys.version_info[0] == 2:
        @gen.coroutine
        def execute_image_operations(self):
            self._override_execute_image_operations()
            yield super(VideoEngineImagingHandler, self).execute_image_operations()
    else:
        exec("\n".join([
            "async def execute_image_operations(self):",
            "    self._override_execute_image_operations()",
            "    await ImagingHandler.execute_image_operations(self)",
        ]))


class ThumborServiceApp(thumbor.app.ThumborServiceApp):
    def get_handlers(self):
        handlers = super(ThumborServiceApp, self).get_handlers()
        for i, handler in list(enumerate(handlers)):
            url_regex, handler_cls, ctx = handler[0], handler[1], handler[2:]
            if issubclass(handler_cls, ImagingHandler):
                # Replace with our custom imaging handler
                handlers[i] = (url_regex, VideoEngineImagingHandler) + ctx
        return handlers



import json
import os
from subprocess import Popen, PIPE

import six
try:
    from shutil import which
except ImportError:
    from thumbor.utils import which

from thumbor_video_engine.exceptions import FFmpegError
from thumbor_video_engine.utils import named_tmp_file


FFPROBE_PATH = os.getenv('FFPROBE_PATH', None)


def ffprobe(buf, extension=None, flat=True):
    """
    Returns a dict based on the json output of ffprobe. If ``flat`` is ``True``,
    the 'format' key-values are made top-level, as well as the first video stream
    in the file (the rest are discarded). Any 'stream' keys that have the same
    name as a key in 'format' are prefixed with ``stream_``.
    """
    global FFPROBE_PATH

    if FFPROBE_PATH is None:
        FFPROBE_PATH = which('ffprobe')

    if FFPROBE_PATH is None:
        raise FFmpegError("Could not find ffprobe executable")

    with named_tmp_file(data=buf, extension=extension) as input_file:
        command = [
            FFPROBE_PATH, '-hide_banner', '-loglevel', 'fatal', '-show_error',
            '-show_format', '-show_streams', '-print_format', 'json',
            '-i', input_file,
        ]

        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=PIPE)

        stdout, stderr = proc.communicate()

        try:
            probe_data = json.loads(stdout)
        except ValueError:
            probe_data = None

        if not isinstance(probe_data, dict):
            raise FFmpegError("ffprobe returned invalid data")

        if 'error' in probe_data:
            raise FFmpegError("%(string)s (%(code)s)" % probe_data['error'])

        if 'format' not in probe_data or 'streams' not in probe_data:
            raise FFmpegError("ffprobe returned invalid data")

        if not flat:
            return probe_data

        try:
            video_stream = next(s for s in probe_data['streams'] if s['codec_type'] == 'video')
        except StopIteration:
            raise FFmpegError("File is missing a video stream")

        data = probe_data['format']
        for k, v in six.iteritems(video_stream):
            if k in data:
                k = 'stream_%s' % k
            data[k] = v
        return data



# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com
# Copyright (c) 2015 Wikimedia Foundation

from thumbor.engines.gif import Engine as BaseEngine
try:
    from shutil import which
except ImportError:
    from thumbor.utils import which


class Engine(BaseEngine):

    def resize(self, width, height):
        width, height = int(width), int(height)
        super(Engine, self).resize(width, height)
        # Allow Gifsicle to add intermediate colors when resizing images.
        # Normally, Gifsicle's resize algorithms use input images' color
        # palettes without changes. When shrinking images with very few colors
        # (e.g., pure black-and-white images), adding intermediate colors can
        # improve the results. The following option allows Gifsicle to add
        # intermediate colors for images that have fewer than 64 input colors.
        self.operations.append("--resize-colors 64")

    def run_gifsicle(self, command):
        if not self.context.server.gifsicle_path:
            gifsicle_path = self.context.config.GIFSICLE_PATH or which('gifsicle')
            if not gifsicle_path:
                raise RuntimeError(
                    "gif engine was requested, but gifsicle binary cannot be found")
            self.context.server.gifsicle_path = gifsicle_path
        if self.context.config.GIFSICLE_ARGS:
            command += " %s" % " ".join(self.context.config.GIFSICLE_ARGS)
        return super(Engine, self).run_gifsicle(command)



from os.path import join
import tc_aws.result_storages.s3_storage
from .base import BaseStorage


class Storage(BaseStorage, tc_aws.result_storages.s3_storage.Storage):

    def _normalize_path(self, path):
        """
        Adapts path based on configuration (root_path for instance)
        :param string path: Path to adapt
        :return: Adapted path
        :rtype: string
        """
        path = path.lstrip('/')  # Remove leading '/'
        path_segments = [path]

        root_path = self._get_config('ROOT_PATH')
        if root_path and root_path != '':
            path_segments.insert(0, root_path)

        transcode_segment = self.get_auto_path_component()

        if transcode_segment:
            path_segments.append(transcode_segment)

        if self._should_randomize_key():
            path_segments.insert(0, self._generate_digest(path_segments))

        if len(path_segments) > 1:
            normalized_path = join(path_segments[0], *path_segments[1:]).lstrip('/')
        else:
            normalized_path = path_segments[0]

        if normalized_path.endswith('/'):
            normalized_path += self.context.config.TC_AWS_ROOT_IMAGE_NAME
        return normalized_path



import sys
from setuptools import setup, find_packages
from os import path

from io import open


CURR_DIR = path.abspath(path.dirname(__file__))


def read(rel_path):
    with open(path.join(CURR_DIR, rel_path), encoding='utf-8') as f:
        return f.read()


def get_long_description():
    readme_rst = read('README.rst')
    lines = readme_rst.split(u"\n")
    return u"\n".join(lines[12:])


def get_version():
    for line in read('src/thumbor_video_engine/__init__.py').splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='thumbor-video-engine',
    version=get_version(),
    description='An engine and tools for manipulating videos with thumbor using ffmpeg',
    long_description=get_long_description(),
    long_description_content_type='text/x-rst',
    url='https://github.com/theatlantic/thumbor-video-engine',
    author='Frankie Dintino',
    author_email='fdintino@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=['thumbor<7' if sys.version_info[0] == '2' else 'thumbor'],
    zip_safe=False)



import hashlib
from os.path import join
from six.moves.urllib.parse import unquote

from thumbor.result_storages import file_storage
from .base import BaseStorage


class Storage(BaseStorage, file_storage.Storage):
    def normalize_path(self, path):
        digest = hashlib.sha1(unquote(path).encode("utf-8")).hexdigest()

        auto_path_component = self.get_auto_path_component()

        return "%s/%s/%s/%s/%s" % (
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH.rstrip('/'),
            "auto_%s" % auto_path_component if auto_path_component else "default",
            digest[:2],
            digest[2:4],
            digest[4:]
        )

    def normalize_path_legacy(self, path):
        try:
            # Python 2
            path = unquote(path).decode('utf-8')
        except AttributeError:
            # Python 3
            path = unquote(path)

        path_segments = [
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH.rstrip("/"),
            file_storage.Storage.PATH_FORMAT_VERSION,
        ]

        auto_path_component = self.get_auto_path_component()
        if auto_path_component:
            path_segments.append(auto_path_component)

        path_segments.extend([self.partition(path), path.lstrip("/")])

        normalized_path = join(*path_segments).replace("http://", "")
        return normalized_path



from __future__ import unicode_literals

from contextlib import contextmanager
import copy
from decimal import Decimal
from io import BytesIO, open
import re
from subprocess import Popen, PIPE

from PIL import Image, ImageSequence
from thumbor.engines import BaseEngine
from thumbor.utils import logger

from thumbor_video_engine.exceptions import FFmpegError
from thumbor_video_engine.ffprobe import ffprobe
from thumbor_video_engine.utils import named_tmp_file, make_tmp_dir, has_transparency


FORMATS = {
    '.mp4': 'mp4',
    '.webm': 'webm',
    '.gif': 'gif',
    '.gifv': 'mp4',
    '.webp': 'webp',
}


class Engine(BaseEngine):

    def __init__(self, context):
        self.original_size = 1, 1
        self.duration = 0
        self.crop_info = 1, 1, 0, 0
        self.image_size = 1, 1
        self.rotate_degrees = 0
        self.flipped_vertically = False
        self.flipped_horizontally = False
        self.resized = False
        self.cropped = False
        self.grayscale = False
        super(Engine, self).__init__(context)
        self.ffmpeg_path = self.context.config.FFMPEG_PATH
        self.ffprobe_path = self.context.config.FFPROBE_PATH

    @property
    def use_gif_engine(self):
        return self.context.config.FFMPEG_USE_GIFSICLE_ENGINE

    @property
    def size(self):
        return self.image_size

    @property
    def source_width(self):
        return self.original_size[0]

    @source_width.setter
    def source_width(self, width):
        orig_w, orig_h = self.original_size
        self.original_size = (width, orig_h)

    @property
    def source_height(self):
        return self.original_size[1]

    @source_height.setter
    def source_height(self, height):
        orig_w, orig_h = self.original_size
        self.original_size = (orig_w, height)

    @property
    def original_size(self):
        return self._original_size

    @original_size.setter
    def original_size(self, value):
        width, height = value
        self._original_size = width, height
        self.crop_info = width, height, 0, 0
        self.image_size = width, height

    def is_multiple(self):
        return False

    def can_convert_to_webp(self):
        """
        We wouldn't want to auto-convert videos to webp, but animated gifs
        should be fair game
        """
        return self.context.config.FFMPEG_GIF_AUTO_WEBP and self.extension == '.gif'

    def resize(self, width, height):
        width, height = int(width), int(height)
        self.resized = True
        self.operations.append(('resize', (width, height)))
        logger.debug('resize {0} {1}'.format(width, height))
        self.image_size = width, height

    def crop(self, left, top, right, bottom):
        left, top, right, bottom = int(left), int(top), int(right), int(bottom)
        self.cropped = True
        self.operations.append(('crop', (left, top, right, bottom)))
        logger.debug('crop {0} {1} {2} {3}'.format(left, top, right, bottom))
        old_out_width, old_out_height, old_left, old_top = self.crop_info
        old_width, old_height = self.image_size

        width = right - left
        height = bottom - top
        self.image_size = width, height

        out_width = int(1.0 * width / old_width * old_out_width)
        out_height = int(1.0 * height / old_height * old_out_height)
        new_left = int(old_left + 1.0 * left / old_width * old_out_width)
        new_top = int(old_top + 1.0 * top / old_height * old_out_height)
        self.crop_info = out_width, out_height, new_left, new_top

    def rotate(self, degrees):
        self.operations.append(('rotate', (degrees,)))
        self.rotate_degrees = degrees

    def flip_vertically(self):
        self.operations.append(('flip_vertically', tuple()))
        self.flipped_vertically = not self.flipped_vertically

    def flip_horizontally(self):
        self.operations.append(('flip_horizontally', tuple()))
        self.flipped_horizontally = not self.flipped_horizontally

    def convert_to_grayscale(self, update_image=True, alpha=True):
        self.operations.append(('convert_to_grayscale', tuple()))
        self.grayscale = True

    # mp4 have no exif data and thus can't be auto oriented
    def reorientate(self, override_exif=True):
        pass

    def load(self, buffer, extension):
        self.extension = extension
        self.buffer = buffer
        self.operations = []
        if self.get_mimetype(buffer).startswith('image/'):
            self.image = Image.open(BytesIO(buffer))
        else:
            # self.image cannot be None, or the thumbor handler returns a 400
            self.image = ''
        self.probe()

    def has_transparency(self):
        if self.image:
            return has_transparency(self.image)
        else:
            return False

    def probe(self):
        # If self.image, we have an animated image (e.g. webp) that ffmpeg
        # cannot yet decode, but pillow can.
        if self.image:
            self.original_size = self.image.size
            duration_ms = 0
            # Load all frames, get the sum of all frames' durations
            for frame in ImageSequence.Iterator(self.image):
                frame.load()
                duration_ms += frame.info['duration']
            self.image.seek(0)
            self.duration = Decimal(duration_ms) / Decimal(1000)
        else:
            ffprobe_data = ffprobe(self.buffer, extension=self.extension)
            self.original_size = ffprobe_data['width'], ffprobe_data['height']
            self.duration = Decimal(ffprobe_data['duration'])

    def read(self, extension=None, quality=None):
        if quality is None:
            return self.buffer  # return the original data
        return self.transcode(extension)

    def transcode(self, extension):
        if getattr(self.context.request, 'format', None):
            out_format = self.context.request.format
            if out_format in ('hevc', 'h264', 'h265'):
                extension = '.mp4'
                self.context.request.format = 'mp4'
        else:
            out_format = FORMATS[extension]

        with self.make_src_file(extension) as src_file:
            if out_format == 'webp':
                return self.transcode_to_webp(src_file)
            elif out_format in ('webm', 'vp9'):
                return self.transcode_to_vp9(src_file)
            elif out_format in ('mp4', 'h264'):
                return self.transcode_to_h264(src_file)
            elif out_format in ('hevc', 'h265'):
                return self.transcode_to_h265(src_file)
            elif out_format == 'gif':
                return self.transcode_to_gif(src_file)
            else:
                raise FFmpegError("Invalid video format '%s' requested" % out_format)

    @contextmanager
    def make_src_file(self, extension):
        mime = self.get_mimetype(self.buffer)
        is_webp = mime == 'image/webp'
        if not is_webp:
            with named_tmp_file(data=self.buffer, suffix=extension) as src_file:
                yield src_file
        else:
            with make_tmp_dir() as tmp_dir:
                im = self.image
                num_frames = im.n_frames
                num_digits = len(str(num_frames))
                format_str = "%(dir)s/%(idx)0{}d.tif".format(num_digits)
                concat_buf = BytesIO()
                concat_buf.write(b"ffconcat version 1.0\n")
                concat_buf.write(b"# %dx%d\n" % im.size)
                for i, frame in enumerate(ImageSequence.Iterator(im)):
                    frame.load()
                    duration_ms = im.info['duration']
                    out_file = format_str % {'dir': tmp_dir, 'idx': i}
                    frame.save(out_file, lossless=True)
                    concat_buf.write(b"file '%s'\n" % out_file.encode('utf-8'))
                    duration_str = "%s" % (Decimal(duration_ms) / Decimal(1000))
                    concat_buf.write(b"duration %s\n" % duration_str.encode('utf-8'))
                self.buffer = concat_buf.getvalue()
                with named_tmp_file(data=self.buffer, suffix='.txt') as src_file:
                    yield src_file

    def get_config(self, prop, format):
        req_val = getattr(self.context.request, prop, None)
        if req_val is not None:
            return req_val
        else:
            config_key = ('ffmpeg_%s_%s' % (format, prop)).upper()
            return getattr(self.context.config, config_key)

    def transcode_to_webp(self, src_file):
        is_lossless = self.get_config('lossless', 'webp')
        if is_lossless or self.has_transparency():
            pix_fmt = 'rgba'
        else:
            pix_fmt = 'yuv420p'

        vf_flags = ['-vf', ','.join(self.ffmpeg_vfilters)] if self.ffmpeg_vfilters else []

        flags = [
            '-loop', '0', '-an', '-pix_fmt', pix_fmt, '-movflags', 'faststart',
        ] + vf_flags + ['-f', 'webp']

        if is_lossless:
            flags += ['-lossless', '1']
        if self.context.config.FFMPEG_WEBP_PRESET:
            flags += ['-preset', "%s" % self.context.config.FFMPEG_WEBP_PRESET]
        if self.context.config.FFMPEG_WEBP_COMPRESSION_LEVEL is not None:
            flags += [
                '-compression_level',
                "%s" % self.context.config.FFMPEG_WEBP_COMPRESSION_LEVEL]
        if self.context.config.FFMPEG_WEBP_QSCALE is not None:
            flags += ['-qscale', "%s" % self.context.config.FFMPEG_WEBP_QSCALE]

        return self.run_ffmpeg(src_file, 'webp', flags=flags, two_pass=False)

    def transcode_to_gif(self, src_file):
        with named_tmp_file(suffix='.png') as palette_file:
            if not self.use_gif_engine and self.ffmpeg_vfilters:
                libav_filter = ','.join(self.ffmpeg_vfilters)
            else:
                libav_filter = 'scale=%d:%d:flags=lanczos' % self.original_size

            input_flags = ['-f', 'concat', '-safe', '0'] if src_file.endswith('.txt') else []

            self.run_cmd([
                self.ffmpeg_path, '-hide_banner',
            ] + input_flags + [
                '-i', src_file,
                '-lavfi', "%s,palettegen" % libav_filter,
                '-y', palette_file,
            ])

            gif_buffer = self.run_cmd([
                self.ffmpeg_path, '-hide_banner',
            ] + input_flags + [
                '-i', src_file,
                '-i', palette_file,
                '-lavfi', "%s[x];[x][1:v]paletteuse" % libav_filter,
                '-f', 'gif',
                '-',
            ])

            if not self.use_gif_engine:
                return gif_buffer
            else:
                gif_engine = self.context.modules.gif_engine
                gif_engine.load(gif_buffer, '.gif')
                gif_engine.operations.append('-O3')

                for op_fn, op_args in self.operations:
                    gif_engine_method = getattr(gif_engine, op_fn)
                    gif_engine_method(*op_args)

            return gif_engine.read()

    @property
    def ffmpeg_vfilters(self):
        vfilters = []
        if self.grayscale:
            vfilters.append('hue=s=0')
        if self.flipped_vertically:
            vfilters.append('vflip')
        if self.flipped_horizontally:
            vfilters.append('hflip')
        if self.rotate_degrees != 0:
            vfilters.append('rotate={0}'.format(self.rotate_degrees))
        if self.cropped:
            vfilters.append('crop={0}'.format(':'.join([str(i) for i in self.crop_info])))
        # scale must be the last one
        if self.resized:
            vfilters.append(
                'scale={0}:flags=lanczos'.format(':'.join([str(s) for s in self.image_size])))
        return vfilters

    def transcode_to_vp9(self, src_file):
        vf_flags = ['-vf', ','.join(self.ffmpeg_vfilters)] if self.ffmpeg_vfilters else []
        flags = [
            '-c:v', 'libvpx-vp9', '-loop', '0', '-an', '-pix_fmt', 'yuv420p',
            '-movflags', 'faststart',
        ] + vf_flags + ['-f', 'webm']

        if self.context.config.FFMPEG_VP9_VBR is not None:
            flags += ['-b:v', "%s" % self.context.config.FFMPEG_VP9_VBR]
        if self.context.config.FFMPEG_VP9_CRF is not None:
            flags += ['-crf', "%s" % self.context.config.FFMPEG_VP9_CRF]
        if self.context.config.FFMPEG_VP9_DEADLINE:
            flags += ['-deadline', "%s" % self.context.config.FFMPEG_VP9_DEADLINE]
        if self.context.config.FFMPEG_VP9_CPU_USED is not None:
            flags += ['-cpu-used', "%s" % self.context.config.FFMPEG_VP9_CPU_USED]
        if self.context.config.FFMPEG_VP9_ROW_MT:
            flags += ['-row-mt', '1']
        if self.get_config('lossless', 'vp9'):
            flags += ['-lossless', '1']
        if self.context.config.FFMPEG_VP9_MAXRATE:
            flags += ['-maxrate', "%s" % self.context.config.FFMPEG_VP9_MAXRATE]
        if self.context.config.FFMPEG_VP9_MINRATE:
            flags += ['-minrate', "%s" % self.context.config.FFMPEG_VP9_MINRATE]

        two_pass = self.context.config.FFMPEG_VP9_TWO_PASS
        return self.run_ffmpeg(src_file, 'webm', flags=flags, two_pass=two_pass)

    def transcode_to_h264(self, src_file):
        width, height = self.image_size
        # libx264 width and height must be divisible by 2
        if width % 2 or height % 2:
            width = (width // 2) * 2
            height = (height // 2) * 2
            self.resize(width, height)

        vf_flags = ['-vf', ','.join(self.ffmpeg_vfilters)] if self.ffmpeg_vfilters else []

        flags = [
            '-c:v', 'libx264', '-an', '-pix_fmt', 'yuv420p', '-movflags', 'faststart',
        ] + vf_flags + ['-f', 'mp4']

        if self.get_config('tune', 'h264'):
            flags += ['-tune', self.get_config('tune', 'h264')]
        if self.context.config.FFMPEG_H264_VBR is not None:
            flags += ['-b:v', "%s" % self.context.config.FFMPEG_H264_VBR]
        if self.context.config.FFMPEG_H264_CRF is not None:
            flags += ['-crf', "%s" % self.context.config.FFMPEG_H264_CRF]
        if self.context.config.FFMPEG_H264_LEVEL:
            flags += ['-level', "%s" % self.context.config.FFMPEG_H264_LEVEL]
        if self.context.config.FFMPEG_H264_PROFILE:
            flags += ['-profile:v', "%s" % self.context.config.FFMPEG_H264_PROFILE]
        if self.context.config.FFMPEG_H264_PRESET:
            flags += ['-preset', "%s" % self.context.config.FFMPEG_H264_PRESET]
        if self.context.config.FFMPEG_H264_BUFSIZE is not None:
            flags += ['-bufsize', "%s" % self.context.config.FFMPEG_H264_BUFSIZE]
        if self.context.config.FFMPEG_H264_MAXRATE:
            flags += ['-maxrate', "%s" % self.context.config.FFMPEG_H264_MAXRATE]
        if self.context.config.FFMPEG_H264_QMIN:
            flags += ['-qmin', "%s" % self.context.config.FFMPEG_H264_QMIN]
        if self.context.config.FFMPEG_H264_QMAX:
            flags += ['-qmax', "%s" % self.context.config.FFMPEG_H264_QMAX]

        two_pass = self.context.config.FFMPEG_H264_TWO_PASS
        return self.run_ffmpeg(src_file, 'mp4', flags=flags, two_pass=two_pass)

    def transcode_to_h265(self, src_file):
        width, height = self.image_size
        # libx265 width and height must be divisible by 2
        if width % 2 or height % 2:
            width = (width // 2) * 2
            height = (height // 2) * 2
            self.resize(width, height)

        vf_flags = ['-vf', ','.join(self.ffmpeg_vfilters)] if self.ffmpeg_vfilters else []

        flags = [
            '-c:v', 'hevc', '-tag:v', 'hvc1', '-an', '-pix_fmt', 'yuv420p',
            '-movflags', 'faststart',
        ] + vf_flags + ['-f', 'mp4']

        x265_params = []

        if self.get_config('tune', 'h265'):
            flags += ['-tune', self.get_config('tune', 'h265')]
        if self.context.config.FFMPEG_H265_VBR is not None:
            flags += ['-b:v', "%s" % self.context.config.FFMPEG_H265_VBR]
        if self.context.config.FFMPEG_H265_CRF is not None:
            flags += ['-crf', "%s" % self.context.config.FFMPEG_H265_CRF]
        if self.context.config.FFMPEG_H265_PROFILE:
            flags += ['-profile:v', "%s" % self.context.config.FFMPEG_H265_PROFILE]
        if self.context.config.FFMPEG_H265_PRESET:
            flags += ['-preset', "%s" % self.context.config.FFMPEG_H265_PRESET]
        if self.context.config.FFMPEG_H265_BUFSIZE is not None:
            x265_params += ["vbv-bufsize=%s" % self.context.config.FFMPEG_H265_BUFSIZE]
        if self.context.config.FFMPEG_H265_MAXRATE:
            x265_params += ["vbv-maxrate=%s" % self.context.config.FFMPEG_H265_MAXRATE]
        if self.context.config.FFMPEG_H265_CRF_MIN:
            x265_params += ["crf-min=%s" % self.context.config.FFMPEG_H265_CRF_MIN]
        if self.context.config.FFMPEG_H265_CRF_MAX:
            x265_params += ["crf-max=%s" % self.context.config.FFMPEG_H265_CRF_MAX]

        flags += ["-x265-params", ":".join(x265_params)]

        two_pass = self.context.config.FFMPEG_H265_TWO_PASS
        return self.run_ffmpeg(src_file, 'mp4', flags=flags, two_pass=two_pass)

    def run_ffmpeg(self, input_file, out_format, flags=None, two_pass=False):
        flags = flags or []

        input_flags = []
        # text files in concat-format require additional input flags
        if input_file.endswith('.txt'):
            input_flags += ['-f', 'concat', '-safe', '0']
            txt_src = self.buffer.decode('utf-8')
            # If all frames have the same duration, set the -r flag to ensure
            # that no frames get dropped
            durations = set(re.findall(r'duration ([\d\.]+)', txt_src))
            if len(durations) == 1:
                duration = list(durations)[0]
                input_flags += ['-r', '1/%s' % duration]
                flags += ['-r', '1/%s' % duration]

        with named_tmp_file(suffix='.%s' % out_format) as out_file:
            if not two_pass:
                self.run_cmd([
                    self.ffmpeg_path, '-hide_banner',
                ] + input_flags + [
                    '-i', input_file,
                ] + flags + ['-y', out_file])
                with open(out_file, mode='rb') as f:
                    return f.read()

            with named_tmp_file(suffix='.log') as passlogfile:
                if '-x265-params' in flags:
                    params_idx = flags.index('-x265-params') + 1
                    if flags[params_idx]:
                        x265_params = flags[params_idx].split(":")
                    else:
                        x265_params = []
                    pass_one_flags = copy.copy(flags)
                    pass_two_flags = copy.copy(flags)
                    pass_one_flags[params_idx] = ":".join(
                        x265_params + ['pass=1', 'stats=%s' % passlogfile])
                    pass_two_flags[params_idx] = ":".join(
                        x265_params + ['pass=2', 'stats=%s' % passlogfile])
                else:
                    pass_one_flags = flags + [
                        '-pass', '1', '-passlogfile', passlogfile]
                    pass_two_flags = flags + [
                        '-pass', '2', '-passlogfile', passlogfile]

                self.run_cmd([
                    self.ffmpeg_path, '-hide_banner',
                ] + input_flags + [
                    '-i', input_file,
                ] + pass_one_flags + ['-y', '/dev/null'])

                self.run_cmd([
                    self.ffmpeg_path, '-hide_banner',
                ] + input_flags + [
                    '-i', input_file,
                ] + pass_two_flags + ['-y', out_file])

                with open(out_file, mode='rb') as f:
                    return f.read()

    def run_cmd(self, command):
        logger.debug("Running `%s`" % " ".join(command))
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=PIPE)
        proc.command = command
        stdout, stderr = proc.communicate()
        logger.debug(stderr)
        if proc.returncode == 0:
            return stdout
        else:
            err_msg = "%s => %s" % (" ".join(command), proc.returncode)
            err_msg += "\n%s" % stderr
            if self.context.request:
                err_msg += "\n%s" % self.context.request.url
            raise FFmpegError(err_msg)
